﻿require 'json'
require 'openssl'
require 'open-uri'
require 'rest-client'
require 'pp' if __FILE__ == $PROGRAM_NAME

class WatsonAPIClient

  VERSION = '0.8.2'

  class << self

    private

    def retrieve_doc(doc_urls)
      apis  = {}

      # Watson API Explorer
      host = doc_urls[/^https?:\/\/[^\/]+/]
      open(doc_urls, Options, &:read).scan(/<a class="swagger-list--item-link" href="\/(.+?)".*?>\s*(.+?)\s*<\/a>/i) do
        begin
          api = {'path'=>doc_urls + $1, 'title'=>$2.sub(/\s*\(.+?\)$/,'')}
          open(api['path'], Options, &:read).scan(/url:\s*'(.+?)'/) do
            api['path'] = host + $1
          end
          apis[api['title']] = api
        rescue OpenURI::HTTPError
        end
      end

      apis
    end

    # for Swagger 2.0
    def listings(apis)
      methods = Hash.new {|h,k| h[k] = {}}
      digest  = Hash.new {|h,k| h[k] = {}}
      apis['paths'].each_pair do |path, operations|
        common_parameters = nil
        operations.each_pair do |access, operation|
          if access == 'parameters'
            common_parameters = operation
            next
          end
          body, query, min, max = nil, [], [], []
          if operation['parameters']
            operation['parameters'] += common_parameters if common_parameters
            (0...operation['parameters'].size).to_a.reverse.each do |index|
              parameter = operation['parameters'][index]
              operation['parameters'][index..index] = apis['parameters'][parameter[parameter.keys.first].split('/').last] if parameter.keys.first == '$ref'
            end
            operation['parameters'].each do |parameter|
              param  = parameter['name']
              body ||= param if parameter['in'] == 'body'
              query << param if parameter['in'] == 'query'
              min   << param if parameter['required']
              max   << param
            end
          end
          access   = access.downcase
          nickname = (operation['operationId'] || path.gsub(/\/\{.+?\}/,'').split('/').last) #.sub(/(.)/) {$1.downcase}
          [nickname, nickname+'_'+access].each do |name|
            methods[name][access]  = {'path'=>path, 'operation'=>operation, 'body'=>body, 'query'=>query, 'min'=>min, 'max'=>max, 'consumes'=>operation['consumes']}
          end
          digest[nickname][access] = {'path'=>path, 'summary'=>operation['summary']}
        end
      end
      {'apis'=>apis, 'methods'=>methods, 'digest'=>digest}
    end
  end

  api_docs = {
    :gateway  => 'https://gateway.watsonplatform.net',
    :doc_base => 'https://watson-api-explorer.mybluemix.net/',
    :ssl_verify_mode => OpenSSL::SSL::VERIFY_NONE
  }
  JSON.parse(ENV['WATSON_API_DOCS'] || '{}').each_pair do |key, value|
    api_docs[key.to_sym] = value
  end
  doc_urls = api_docs.delete(:doc_base)
  Gateways = api_docs.delete(:gateway)
  Options  = api_docs
  Services = JSON.parse(ENV['VCAP_SERVICES'] || '{}')
  DefaultParams = {:user=>'username', :password=>'password'}
  AvailableAPIs = []

  retrieve_doc(doc_urls).each_value do |list|
    AvailableAPIs << list['title'].gsub(/\s+(.)/) {$1.upcase}
    env = list['title'].gsub(/\s+/,'_').downcase
    module_eval %Q{
      class #{list['title'].gsub(/\s+(.)/) {$1.upcase}} < WatsonAPIClient
        Service = WatsonAPIClient::Services['#{env}']
        RawDoc  = "#{list['path']}"

        class << self
          alias :_const_missing :const_missing

          def const_missing(constant)
            if constant == :API
              const_set(:API, listings(JSON.parse(open(RawDoc, WatsonAPIClient::Options, &:read))))
            else
              _const_missing(constant)
            end
          end
        end
        pp [self, 'See ' + RawDoc, API['digest']] if '#{__FILE__}' == '#{$PROGRAM_NAME}'
      end
    }
  end

  # All subclass constructors use following hash parameter - 
  #
  # @param [Hash] options See following..
  # @option options [String] :url                API URL (default: the url described in listings or VCAP_SERVICES)
  # @option options [String] :user               USER ID (default: the username described in VCAP_SERVICES)
  # @option options [String] :password           USER Password (default: the password described in VCAP_SERVICES)
  # @option options [String] :apikey or :api_key API key for the Alchemy API (default: the apikey or api_key described in VCAP_SERVICES)
  # @option options [String] :version            The release date of the version of the Alchemy API.
  # @option options [Object] other_options       Other options are passed to RestClient::Resource.new[http://www.rubydoc.info/gems/rest-client/RestClient/Resource] as it is. 
  #
  # @note VCAP_SERVICES[http://www.ibm.com/watson/developercloud/doc/getting_started/gs-variables.shtml#vcapServices] is IBM Bluemix™ environment variable.
  #
  def initialize(options={})
    define_api_methods
    set_variables(options)
    @url   ||= Gateways + self.class::API['apis']['basePath']
    @options = {}
    self.class.superclass::DefaultParams.each_pair do |sym, key|
      @options[sym] = @credential[key] if @credential.key?(key)
    end
    @options.update(options)
    @service = RestClient::Resource.new(@url, @options)
  end

  private

  def define_api_methods
    self.class::API['methods'].each_pair do |method, definition|
      self.class.module_eval %Q{define_method("#{method}",
        Proc.new {|options={}|
          begin
            rest_access_#{definition.keys.size > 1                  ? 'auto_detect' :
                          definition[definition.keys.first]['body'] ? 'with_body'   :
                                                                      'without_body' }("#{method}", options.dup)
          rescue RestClient::ExceptionWithResponse => error
            STDERR.puts ["Failed method: #{self.class}##{method}", options]
            raise error
          end
        }
      )} unless respond_to?(method)
    end
  end

  def set_variables(options)
    @credential = self.class::Service ? self.class::Service.first['credentials'] : {}
    if options.key?(:url)
      @url = options.delete(:url)
    elsif @credential.key?('url')
      @url = @credential['url']
    end
  end

  def rest_access_auto_detect(method, options)
    definition = self.class::API['methods'][method.to_s]
    options[:access] ||= select_access(definition, options)
    definition[options[:access]]['body'] ?
      rest_access_with_body(method, options)  :
      rest_access_without_body(method, options)
  end

  def rest_access_without_body(method, options)
    path, access = swagger_info(method, options)
    options = {:params => options} if access == 'get'
    @service[path].send(access, options)
  end

  def rest_access_with_body(method, options)
    path, access, spec = swagger_info(method, options)
    body = options.delete(spec['body'])
    @service[path].send(access, body, options)
  end

  def select_access(definition, options)
    definition.keys.reverse.each do |access|
      spec = definition[access]
      keys = options.keys.map {|key| key.to_s}
      return access if (keys - spec['max']).empty? && (spec['min'] - keys).empty?
    end
    raise ArgumentError, "Cannot select the suitable access method from '#{definition.keys.join(', ')}', see #{self.class::RawDoc}."
  end

  def swagger_info(method, options)
    definition = self.class::API['methods'][method.to_s]
    access = (options.delete(:access) || definition.keys.first).downcase
    spec   = definition[access]
    options.keys.each do |key|
      options[key.to_s] = options.delete(key) if key.kind_of?(Symbol)
    end
    (spec['min'] - options.keys).each do |key|
      options[key.to_s] = @options[key.to_sym] if @options.include?(key.to_sym)
    end
    lacked = spec['min'] - options.keys
    extra  = options.keys - spec['max']
    raise ArgumentError, "Lacked parameter(s) : '#{lacked.join(', ')}', see #{self.class::RawDoc}." unless lacked.empty?
    raise ArgumentError, "Extra parameter(s) : '#{extra.join(', ')}', see #{self.class::RawDoc}."   unless extra.empty?
    query  = {}
    spec['query'].each do |param|
      query[param] = options.delete(param) if options.key?(param)
    end
    options[:content_type] = spec['consumes'].first if spec['consumes'].kind_of?(Array) && spec['consumes'].length == 1
    path  = spec['path'].gsub(/\{(.+?)\}/) {options.delete($1)}
    path += '?' + URI.encode_www_form(query) unless query.empty?
    [path, access, spec]
  end
end
